#ifndef PTR_PARTAGE_H
#define PTR_PARTAGE_H

#include "Partageable.h"
#include "EstConvertible.h"
#include "Clonable.h"
#include <utility> // std::swap()

template <class T>
   class PtrPartage
   {
      static_assert(EstConvertible<T*,Partageable*>::VAL, "Le type T doit tre partageable");
      T *ptr_;
      class AutoRelacheur
      {
         T *tmpptr_;
      public:
         AutoRelacheur(T *p)
            : tmpptr_(p)
         {
            if (tmpptr_) tmpptr_->AddRef();
         }
         ~AutoRelacheur() throw()
            { if (tmpptr_) tmpptr_->Release(); }
      };
   public:
      explicit PtrPartage(T *p = 0)
         : ptr_(p)
      {
         if (ptr_) ptr_->AddRef();
      }
      PtrPartage(const PtrPartage &p)
         : ptr_(p.ptr_)
      {
         if (ptr_) ptr_->AddRef();
      }
      T *get() const throw()
         { return ptr_; }
      //
      // Ici, on est forcs de duplique le point. Prudence
      //
      template <class U>
         PtrPartage(const PtrPartage<U> &p)
            : ptr_(0)
         {
            if (p.get())
            {
               ptr_ = dynamic_cast<T*>(p.get());
               ptr_->AddRef();
            }
         }
      void swap (PtrPartage &p)
         { std::swap (ptr_, p.ptr_); }
      PtrPartage& operator=(const PtrPartage &p)
      {
         PtrPartage(p).swap(*this);
         return *this;
      }
      template <class U>
         PtrPartage& operator=(const PtrPartage<U> &p)
         {
            PtrPartage (p).swap(*this);
            return *this;
         }
      ~PtrPartage() throw()
      {
         if (ptr_) ptr_->Release();
      }
      T& operator*()
         { return ptr_; }
      const T& operator*() const // indirection importante si T est abstrait
         { return ptr_; }
      T* operator->()
      {
         AutoRelacheur ar(ptr_);
         return ptr_;
      }
      const T* operator->() const
      {
         AutoRelacheur ar(ptr_);
         return ptr_;
      }
      //
      // RAFFINER
      //
      operator bool() const
         { return get() != 0; }
   };

template <class T, class U>
   PtrPartage<T> partageable_cast(const PtrPartage<U> &p)
   {
      return PtrPartage<T>(p);
   }

#endif

